	/*

	Copyright (C) 1998 Stefan Westerfeld
                       stefan@space.twc.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    */

#include <sys/types.h>
#include <sys/time.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include "synth_impl.h"
#include "resources.h"
#include "midi.h"
#include "midibus.h"
#include "structures.h"
#include "ioevent.h"
#include "synth_server.h"
#include "debug.h"
 
extern "C" {
#include "sound.h"
#include "utils.h"
}

#define Die(a) { fprintf(stderr,"ERROR: %s\n",a); exit(1); }

//****************** Synthesizer *******************

Synthesizer_impl::Synthesizer_impl()
{
	Arts::RealtimeConfig InitRTConfig;

	describe("Synthesizer");

	InitRTConfig.Cycles = 128;		// calculate 128 cycles at once by default
	InitRTConfig.CacheSizeKB = 8192;	// 8 megs of cache are reasonable
	InitRTConfig.RingBufferSize = 64;	// ringbuffer size (2^n)
	InitRTConfig.Debug = false;		// no debugging by default

	RTConfig(InitRTConfig);

	artsdebug("... Synthesizer_impl created\n");
	SynthModules = NULL;
	_Cache = new Cache;
	MCount = 0;
	_StructChangeCount = 0;
	MS_SynthModule = (ModuleServer<SynthModule> *)SynthModule::get_MS();
	nextID = 1;

	inScheduler = false;

	_MidiChannel = new MidiChannel_impl();
}

void Synthesizer_impl::cleanUp()
{
	unsigned long i;

	// release published structures when being shut down
	for(i=0;i<_Structures.length();i++)
		_Structures[i]->decRef();
}

MidiChannel_impl *Synthesizer_impl::getMidiChannel()
{
	return _MidiChannel;
}

long Synthesizer_impl::StructChangeCount()
{
	return _StructChangeCount;
}

Synthesizer_impl::~Synthesizer_impl()
{
	CORBA::release(_MidiChannel);
	artsdebug("... Synthesizer_impl destroyed\n");
}

void Synthesizer_impl::RTConfig(const Arts::RealtimeConfig &newconfig)
{
	assert(newconfig.Cycles > 0);
	_RTConfig = newconfig;
	setartsdebug(_RTConfig.Debug);
}

Arts::RealtimeConfig Synthesizer_impl::RTConfig()
{
	return _RTConfig;
}

Cache *Synthesizer_impl::getCache()
{
	return _Cache;
}

bool Synthesizer_impl::Exec()
{
/** old style (one sample) scheduling **
	long i,cycles = _RTConfig.Cycles;
	while(cycles--)
	{
		for(i=0;i<MCount;i++) SynthModules[i]->Calculate();
	}
****************************************/

/** new style (dynamic size) scheduling **/

	long cycles = _RTConfig.Cycles;
	unsigned long *done = calloc(MCount,sizeof(unsigned long));
	long i,incomplete, died = 0;

	//printf("entering; cycles = %d\n",cycles);
	do {
		incomplete = 0;		/* assume we have calculated all cycles for all
								consumers, and later increment if some are
								still missing */
		for(i=0;i<MCount;i++)
		{
			if(SynthModules[i]->clients == 0)
			{
				//printf("consumber = %s,done = %d, cycles = %d\n",SynthModules[i]->getClassName(),done[i],cycles);
				/* a module whose input is not comsumed from other modules
					is a "push delivery" style module, such as speakers,
					or writing audio to log file, etc. and has to get
					external requests from the scheduling system */

				if(done[i] != cycles)
					done[i] += SynthModules[i]->request(cycles-done[i]);
				assert(done[i] <= cycles);
				if(done[i] != cycles) incomplete++;
				//printf("*scheduler*\n");
				died ++;
				if(died > 10000)
				{
					free(done);
					return false;
				}
			}
		}
	} while(incomplete);

	//printf("<=> done!!\n");
	free(done);
	return true;
}

void Synthesizer_impl::Run()
{
	inScheduler = true;
	bool eresult = Exec();
	inScheduler = false;

	if(!eresult)
	{
		status_halted = true;
		fprintf(stderr,"seems to be uncalculatable?\n");
	}

	while(killList.size())
	{
		long id = *killList.begin();
		freeStructure(id);
		killList.erase(killList.begin());
	}

	if(status_halted)
	{
		usleep(10000);
		return;
	}

	//fprintf(stderr,"Using %ld modules for synthesis\n",MCount);

	// Calculate cpu usage.
	status_cycles += _RTConfig.Cycles;
	if(status_cycles > 20000)
	{
		// do cache cleanups periodically
		status_cache_usage = _Cache->cleanUp(_RTConfig.CacheSizeKB*1024);

	    float cpu_time = (clock()-status_clocks)/(float)CLOCKS_PER_SEC;
		float real_time = benchmark(BENCH_END);

		if(cpu_time > 0 && real_time > 0) // there may be wraparounds
		{
			status_cpu_usage = cpu_time / real_time;
			if(status_cpu_usage > 0.9)	  // more than 90%  -> not good!
			{
				status_stalled ++;
			}
			else status_stalled = 0;

			if(status_stalled > 5)
			{
				// ok, cancel synthesis due to cpu overload!
				status_halted = true;
			}
		}

		// prepare for next checkpoint
		status_clocks = clock(); 
		status_cycles -= 20000;
		benchmark(BENCH_BEGIN);
	}
}

Arts::Status Synthesizer_impl::getStatus()
{
	Arts::Status status;

	status.cpu_usage = status_cpu_usage;
	status.cache_usage = status_cache_usage;
	status.halted = status_halted;
	status.module_count = MCount;
	return(status);
}

void Synthesizer_impl::Reset()
{
	status_halted = false;
	status_stalled = 0;
	status_cpu_usage = 0;
	status_clocks = clock();
	status_cycles = 0;	
	benchmark(BENCH_BEGIN);
}

CORBA::Long Synthesizer_impl::getModuleCount()
{
	return(MS_SynthModule->getModuleCount());
}

void collectPorts(Arts::StringSeq &ports,SynthModule *m,SynthModule::ArgType argtype)
{
	int i = 0;
	string argname;

	do
	{
		argname = m->getArgument(i,argtype);
		if(argname != "")
		{
			i++;
			ports.length(i);
			//printf("%s/inarg[%d] = %s\n",name,i-1,argname.data());
			ports[i-1] = CORBA::string_dup(argname.c_str());
		}
	} while(argname != "");
}

CORBA::Boolean Synthesizer_impl::getModuleInfo( CORBA::Long nr, Arts::ModuleInfo*& info )
{
	char *name=MS_SynthModule->getModuleName(nr);
	if(name)
	{
		SynthModule *m = (SynthModule *)MS_SynthModule->getModule(name);

		if(m)
		{
			info = new Arts::ModuleInfo;
			info->name = CORBA::string_dup(name);

			// collect inports & outports
			collectPorts(info->inports,m,SynthModule::arg_in);
			collectPorts(info->outports,m,SynthModule::arg_out);
			return(true);
		}
	}
	return(false);
}

Arts::StructureDesc *Synthesizer_impl::createStructureDesc()
{
	artsdebug("creating new structure desc\n");

	Arts::StructureDesc_ptr sd = new StructureDesc_impl;
	return Arts::StructureDesc::_duplicate(sd);
}

CORBA::Long Synthesizer_impl::createStructure(Arts::StructureDesc
														*structuredesc)
{
	// no interfaces to the outside world
	//
	// this is just for using the same code for creating structures
	// that are connected with interfaces to the outside world and
	// creating self contained structures
	list<SynthModule *> interfaces;

	return createConnectedStructure(structuredesc,interfaces);
}

long Synthesizer_impl::createConnectedStructure(Arts::StructureDesc
							*structuredesc, list<SynthModule*> interfaces)
{
	assert(!inScheduler);

	Arts::ModuleDescSeq_var modules = structuredesc->Modules();
	Arts::ModuleDesc *md;
	Arts::PortDesc *pd;
	SynthModule *NewModules[modules->length()];
	string textargs[modules->length()];

	long RBSize = _RTConfig.RingBufferSize;

	_StructChangeCount++;

	artsdebug("Execute I <Creation>\n");
	// first step: create modules, allocate output buffers
	unsigned long m,m2,p;

	for(m=0;m<modules->length();m++)
	{
		md = (*modules)[m];
		CORBA::String_var modname = md->Name();
		Arts::PortDescSeq_var ports = md->Ports();
		artsdebug("creating module %ld %s\n",m,(char *)modname);

		SynthModule *NewModule;

		if(md->isInterface())
		{
			list<SynthModule *>::iterator iface = interfaces.begin();
			bool found = false;
			artsdebug(" => setting up interface connection\n");

			while(!found && iface != interfaces.end())
			{
				artsdebug("[%s] <=> [%s]\n",(*iface)->getClassName(),(char *)modname);
				if(strcmp((*iface)->getClassName(),modname) == 0)
				{
					found = true;
					NewModule = (*iface);

					// only use once of course!
					interfaces.remove(NewModule);
				}
				iface++;
			}
			assert(found);
		}	
		else
		{
			NewModule = (SynthModule *)MS_SynthModule->getModule(modname);
		}
		assert(NewModule);
		NewModule->Synthesizer = this;

		NewModules[m] = NewModule;
		NewModule->ID = md->ID();
		NewModule->executionID = nextID;

		for(p=0;p<ports->length();p++)
		{
			pd = (*ports)[p];

			if(pd->Type().Direction == Arts::output)
			{
				artsdebug("creating port id %ld\n",pd->ID());

				SynthConnection *conn = new SynthConnection(NewModule,pd->ID());
				conn->allocBuffer(0.0,RBSize);
				/*conn->buffer = new SynthBuffer(0.0,RBSize);*/
				NewModule->assignOutConn(conn);
			}
		}
	}

	// second step: finialize modules ->  connect or allocate input buffers

	artsdebug("Execute II <Connection>\n");
	for(m=0;m<modules->length();m++)
	{
		textargs[m] = "";
		SynthModule *NewModule = NewModules[m];
		md = (*modules)[m];
		Arts::PortDescSeq_var ports = md->Ports();

		for(p=0;p<ports->length();p++)
		{
			Arts::PortType pt;

			pd = (*ports)[p];
			pt = pd->Type();

			if(pt.Direction == Arts::input)
			{
				assert(pt.ConnType != Arts::event);   // no event channels yet

				if(pt.ConnType == Arts::stream)
				{
					assert(pt.DataType == Arts::audio_data);
					
					SynthConnection *conn =
						new SynthConnection(NewModule,pd->ID());
					NewModule->assignInConn(conn);

					if(pd->isConnected())
					{
						Arts::PortDescSeq_var conns = pd->Connections();
						assert(conns->length() == 1);
						Arts::PortDesc *pd2 = (*conns)[0];

						long wantedID = pd2->ID();
						long foundnr = 0;
						SynthModule *other,*found=0;
						// find other module
				
						for(m2=0;m2<modules->length();m2++)
						{
							long p2;
							other = NewModules[m2];
							for(p2=0;p2<other->outConnCount;p2++)
							{
								//if(other->outID[p2] == wantedID)
								if(other->outConn[p2]->ID == wantedID)
								{
									found = other;
									foundnr=p2;
								}
							}
						}
						if(!found)
						{
							CORBA::String_var modname = md->Name();
							artsdebug("fatal: didn't find port %ld, which should provide input for module %s\n", wantedID,(char *)modname);
						}
						assert(found);
						conn->connectToSource(found->outConn[foundnr]);
					}
					else
					{

						float initialvalue = 0.0;

						if(pd->hasValue())
							initialvalue = pd->FloatValue();

						conn->allocBuffer(initialvalue,RBSize);
						//conn->buffer = new SynthBuffer(initialvalue,RBSize);
					}
				}
				if(pt.ConnType == Arts::property)
				{
					assert(pt.DataType == Arts::string_data);
					assert(textargs[m] == "");

					// only one textarg may be passed in that primitive form
					CORBA::String_var arg = pd->StringValue();
					textargs[m] = arg;
					artsdebug("got stringproperty; %s\n",textargs[m].c_str());
				}
			}
		}
	}

	artsdebug("Execute III <prepare for Execution & Initialization>\n");
	for(m=0;m<modules->length();m++)
	{
		SynthModule *NewModule = NewModules[m];
		md = (*modules)[m];

		artsdebug("Initialize: textargs = %s\n",textargs[m].c_str());
		artsdebug("NewModule %s characteristics: in %ld, out %ld\n",NewModule->getClassName(),NewModule->inConnCount,NewModule->outConnCount);
		NewModule->Initialize(textargs[m]);

		// EXPERIMENTAL: prepareExecution after Initialize
		NewModule->prepareExecution();

		if(!md->isInterface())
		{
			SynthModules = (SynthModule **)realloc(SynthModules,
									sizeof(SynthModule *)*(MCount+1));
			SynthModules[MCount++] = NewModule;
		}
	}

	artsdebug("Execute IV <IO Initialization>\n");
	for(m=0;m<modules->length();m++)
	{
		SynthModule *NewModule = NewModules[m];
		int infd,outfd;
		ArtsIOEvent *ev;

		NewModule->getfds(infd,outfd);

		if(infd != -1)
		{
			ev = new ArtsIOEvent(this);
			ArtsOrb->dispatcher()->rd_event(ev,infd);
			ev->executionID = NewModule->executionID;
			IOEvents.push_back(ev);
		}
		if(outfd != -1)
		{
			ev = new ArtsIOEvent(this);
			ArtsOrb->dispatcher()->wr_event(ev,outfd);
			ev->executionID = NewModule->executionID;
			IOEvents.push_back(ev);
		}
	}
	artsdebug("Execute ok\n");
	return nextID++;
}

void Synthesizer_impl::addDynamicConnection (SynthModule *target, SynthModule *source, unsigned long size)
{
	assert(!inScheduler);
	unsigned long i;
	artsdebug(">>>>>>>>>> ADC: target is %s source is %s\n",target->getClassName(),source->getClassName());
	artsdebug(">>>>>>>>>> ADC: source->inConnCount = %d\n",source->inConnCount);
	artsdebug(">>>>>>>>>> ADC: target->inConnCount = %d\n",target->inConnCount);
	artsdebug(">>>>>>>>>> ADC: add channels: %d\n",size);
	assert(source->inConnCount >= size);
	for(i=0;i < size;i++)
	{
		SynthConnection *newconn = new SynthConnection(target,0);
		newconn->isDynamic = true;
		target->assignInConn(newconn);
		// FIXME: will fail when constant stuff
		newconn->connectToSource(source->inConn[i]->sourceconn);
	}
	artsdebug(">>>>>>>>>> after target->inConnCount = %d\n",target->inConnCount);
}

void Synthesizer_impl::removeDynamicConnections (SynthModule *module)
{
	assert(!inScheduler);

	unsigned long i,remove=0;

	for(i=0;i<module->inConnCount;i++)
	{
		if(module->inConn[i]->isDynamic)
		{
			remove++;
			if(module->inConn[i]->sourceconn)
				module->inConn[i]->disconnectFromSource(module->inConn[i]->sourceconn);
			delete module->inConn[i];
		}
		else
		{
			assert(remove == 0);		// dynamic channels should ONLY be attached at the end!
		}
	}
	module->inConnCount -= remove;

	module->inConn = realloc(module->inConn,sizeof(SynthConnection *)*module->inConnCount);
	// FIXME: free memory here (?)
}

void Synthesizer_impl::kill(long id)
{
	// Is called by Synth_STRUCT_KILL
	//
	// As freeing while a scheduling cycle is done is not possible, we push that request to
	// a list and execute the kills after scheduling

	killList.push_back(id);
}

CORBA::Boolean Synthesizer_impl::isExecuting(CORBA::Long id)
{
	if(SynthModules)
	{
		int i;
		for(i=0;i<MCount;i++)
		{
			if(SynthModules[i]->executionID == id) return true;
		}
	}
	return false;
}

CORBA::Boolean Synthesizer_impl::freeStructure(CORBA::Long id)
{
	assert(!inScheduler);

	long NCount = 0,freed = 0,i;
	artsdebug("freeing structure id = %ld\n",id);

	// Should notify the parent that the structure it is executing
	// no longer exists. Currently there is only polling to check that.

	_StructChangeCount++;
	list<ArtsIOEvent *>::iterator ioevent;
	for (ioevent=IOEvents.begin(); ioevent != IOEvents.end(); ioevent++)
	{
		ArtsIOEvent *ev = (*ioevent);
		if(ev->executionID == id)
		{
			ArtsOrb->dispatcher()->remove(ev,CORBA::Dispatcher::All);
			delete ev;
			IOEvents.erase(ioevent);

			// and restart from the beginning because our iterator could
			// now point to nirwana
			ioevent = IOEvents.begin();

			artsdebug("removed io event\n");
		}
	}

	if(SynthModules)
	{
		// kill those that match the id

		// 1st step: deinitialize them
		for(i=0;i<MCount;i++)
		{
			if(SynthModules[i]->executionID == id) SynthModules[i]->DeInitialize();
		}

		// 2nd step: delete them
		for(i=0;i<MCount;i++)
		{
			if(SynthModules[i]->executionID == id)
			{
				delete SynthModules[i];
				freed++;
			}
			else SynthModules[NCount++] = SynthModules[i]; // "compress" modules
		}

		MCount = NCount;
		SynthModules = (SynthModule **)realloc(SynthModules,
									sizeof(SynthModule *)*MCount);
		if(MCount == 0) SynthModules = 0;

		if(freed)
		{
			artsdebug("freed %ld modules\n",freed);
			return(true);
		}
	}
	artsdebug("no modules freed\n");
	return false;
}

void Synthesizer_impl::publishStructureDesc(Arts::StructureDesc *structuredesc)
{
	Arts::StructureDesc_ptr structure;

	unsigned long l = 0;
	bool found = false;

	// create deep copy by saving and reloading
	Arts::StringSeq_var strings = structuredesc->saveToList();

	// reload below

	while(l < _Structures.length() && !found)
	{
		CORBA::String_var name1=_Structures[l]->Name();
		CORBA::String_var name2=structuredesc->Name();

		found = !strcmp(name1,name2);
		if(!found) l++;
	}

	if(!found) // didn't find a structure with the same name
	{
		_Structures.length(l+1);

		structure = createStructureDesc();
		_Structures[l] = structure;
		structure->loadFromList(strings);
		structure->incRef();
	}
	else		// if we did, just overwrite this one
	{
		structure = _Structures[l];
		structure->clear();		// clear old contents
		structure->loadFromList(strings);
	}
}

Arts::StructureDesc *Synthesizer_impl::lookupStructureDesc(const char *name)
{
	unsigned long i;

	for(i=0;i<_Structures.length();i++)
	{
		CORBA::String_var aname = _Structures[i]->Name();
		if(strcmp(aname,name) == 0)
			return Arts::StructureDesc::_duplicate(_Structures[i]);
	}

	return 0;
}

Arts::StructureDescSeq *Synthesizer_impl::publishedStructures()
{
	return(new Arts::StructureDescSeq(_Structures));
}

Arts::ResourceInfoSeq *Synthesizer_impl::Resources()
{
	Arts::ResourceInfoSeq *resources = new Arts::ResourceInfoSeq;
	list<ResourceMonitor *> &rlist = *ResourceList::the()->getResources();
	list<ResourceMonitor *>::iterator i;
	unsigned long nr = 0;

	resources->length(rlist.size());
	for(i=rlist.begin();i != rlist.end();i++)
	{
		Arts::ResourceInfo& ri = (*resources)[nr++];
		ri.Used = (*i)->Used();
		ri.Description = CORBA::string_dup((*i)->Description());
	}
	return(resources);
}
